clc;clear;
% This simple script contains many examples of FFW package.
%
% NOTE:
% the optimized parameters for FFW algorithm depend only on
% sizes of input matrices and on their values (real or complex).
% If the same filter has to be applied to several images, its FFT2
% value can be determined only one time, saving computational time.
% In this case (the same filter applied to several images) the
% determination of optimized parameter must not take into account the
% computational cost of such operation, since it is done only once.
%
% Luigi Rosa
% Via Centrale 35
% 67042 Civita Di Bagno
% L'Aquila ITALY
% mobile:   +39 3207214179
% email:    luigi.rosa@tiscali.it
% web site: http://www.advancedsourcecode.com
%

% Load execution times for:
% FFTrv      FFT execution times for real 1D vectors
% FFTiv      FFT execution times for complex 1D vectors
% IFFTsym    IFFT execution times for complex 1D vectors with option
%            'symmetric' enabled (i.e. real output)
% IFFTnonsym IFFT execution times for complex 1D vectors with option
%            'nonsymmetric' enabled (i.e. complex output)
%
% These times have been calculated with the script papiclock.m 
% from length N = 3 up to length N = 2048.
% A finer determination of such times have been done using PAPI for Matlab
% http://www.mathworks.com/matlabcentral/fileexchange/loadFile.do?objectId=5445&objectType=File
% or http://icl.cs.utk.edu/papi/

if exist('papifftexecutiontimes.mat')>0
    load papifftexecutiontimes
else
    disp('The file papifftexecutiontimes.mat (collection of vectors with execution times');
    disp('of FFT and IFFT operators) does not exist. This file has to be created using');
    disp('the script papiclock.m that requires PAPI for Matlab installed.');
end
%--------------------------------------------------------------------------
%------------------------------------------------------ FAST 2D CONVOLUTION
disp('-----------------------------------------------------------------');
disp('***   FFW algorithm - 2D convolution   ***');
disp(' ');
% Two arbitrary 2D matrices
a = rand(194,182);
b = rand(192,173);
% Add a complex part
% a=a+i*rand(size(a));
% b=b+i*rand(size(b));
 
% Optimized parameters for 2D convolution
opt = dbl(FFTrv,FFTiv,IFFTsym,IFFTnonsym,size(a),size(b),isreal(a),isreal(b));

tic;
y0 = fftolamopt2(a,b,opt);
t0 = toc;
% Another example:
% y1 = fftolamopt2(a,b,opt,'same');
% equivalent to do y1 = conv2(a,b,'same');

tic;
y1 = conv2(a,b,'full');
t1 = toc;

tic;
size_out = size(a)+size(b)-1;
y2       = ifft2(fft2(a,size_out(1),size_out(2)).*fft2(b,size_out(1),size_out(2)));
if isreal(a) && isreal(b)
    y2 = real(y2);
end
t2 = toc;

disp('FFW algorithm - Time required');
disp(t0);
disp('Conv2 - Time required');
disp(t1);
disp('Standard FFT based convolution - Time required');
disp(t2);
disp('Max abs error');
disp(max(max(abs(y0-y1))));
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%-------------------------------------------------------- FAST 2D FILTERING
disp('-----------------------------------------------------------------');
disp('***   FFW algorithm - 2D filtering   ***');
disp(' ');
I = imread('board.tif');
I = rgb2gray(I);
I = double(I);

% If the kernel filter is too small it is convenient to work in time domain, 
% without any FFT!
myfilter = rand(60,50);
tic;
If       = filter2(myfilter,I,'same');
t0       = toc;

% Rot90 of myfilter to make results consistent
mf       = fliplr(flipud(myfilter));
% equivalent to do mf = rot90(myfilter);

% Optimized parameters for 2D filtering
opt      = dbl(FFTrv,FFTiv,IFFTsym,IFFTnonsym,size(I),size(mf),isreal(I),isreal(mf));
tic;
If1      = fftolamopt2(I,mf,opt,'same');
t1       = toc;


disp('FFW algorithm - Time required');
disp(t1);
disp('Filter2 - Time required');
disp(t0);
disp('Max abs error');
disp(max(max(abs(If-If1))));
%--------------------------------------------------------------------------
%---------------------------------------- FAST NORMALIZED CROSS-CORRELATION
% This function can easily be used also for a fast 
% normalized cross-correlation.
% See normxcorr2 function (Matlab Image Processing Toolbox) for more details.
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------

%--------------------------------------------------------------------------
%------------------------------------------------------- FINE TUNING OPTION
disp('-----------------------------------------------------------------');
disp('***   FFW algorithm - Fine tuning option   ***');
disp(' ');
a = rand(234,321);
b = rand(888,654);
% FFW algorithm with fine tuning option NOT enabled
finetuning = 0;
opt        = dbl(FFTrv,FFTiv,IFFTsym,IFFTnonsym,size(a),size(b),isreal(a),isreal(b),finetuning);

tic;
y0  = fftolamopt2(a,b,opt);
t   = toc;
disp('FFW algorithm - Time required: fine tuning option NOT enabled');
disp(t);
% FFW algorithm with fine tuning option enabled
finetuning = 1;
opt        = dbl(FFTrv,FFTiv,IFFTsym,IFFTnonsym,size(a),size(b),isreal(a),isreal(b),finetuning);

tic;
y0  = fftolamopt2(a,b,opt);
t   = toc;
disp('FFW algorithm - Time required: fine tuning option enabled');
disp(t);
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------




